import pandas as pd
import aiohttp
import requests
import asyncio
from time import time, sleep
from os import environ
from globy_webcrawlers.website import Website
from globy_webcrawlers.common import Logger, domain_from_url
from fake_useragent import UserAgent
import gzip
from io import BytesIO
import brotli
import zlib
from pathlib import Path
from enum import Enum
from globy_webcrawlers.scrapy_backend import run_globy_spider

class Backend(Enum):
    GLOBY = 1
    SCRAPY = 2

class WebSiteDataCrawler:
    """
    Loads URLs and scrapes (a)synchronously for data
    """

    def __init__(self, debug=False, store_html_to_file=False, backend="globy") -> None:
        self.logger = Logger().logger
        self.urls = []
        self.dataframes = []
        self.limit_websites = 100
        self.output_file = None
        self.simulate_browser = True
        self.store_html_to_file = store_html_to_file
        self.html_output_folder = "html_output"
        if self.store_html_to_file:
            Path(self.html_output_folder).mkdir(parents=True, exist_ok=True)

        # Timeout in seconds to use per website
        self.crawl_timeout = 30

        # Used for debugging
        self._latest_url = None
        self._latest_response = None
        self.debug = debug
        if self.debug:
            print("Debugging/inspection enabled. All responses will be stored in WebSiteDataCrawler.debugging_data as (url, response) tuples")
            environ['PYTHONINSPECT'] = 'TRUE'
        self.all_websites = []
        self.limit_urls()

        # Select backend
        if backend == "scrapy":
            self.backend = Backend.SCRAPY
        else:
            self.backend = Backend.GLOBY

    
    def limit_urls(self):
        if len(self.urls) > self.limit_websites:
            self.urls = self.urls[:self.limit_websites]
            print(f"<> Limiting URLs to {self.limit_websites}. If you wish to increase this limit, set WebSiteDataCrawler.limit_websites to a higher value. <>")

    def decode_compressed_response(self, response):
        try:
            if 'Content-Encoding' in response.headers:
                content_encoding = response.headers['Content-Encoding'].lower()

                if 'gzip' in content_encoding:
                    compressed_data = response.content
                    buffer = BytesIO(compressed_data)
                    with gzip.GzipFile(fileobj=buffer, mode='rb') as f:
                        decompressed_data = f.read()
                elif 'deflate' in content_encoding:
                    compressed_data = response.content
                    decompressed_data = zlib.decompress(compressed_data, -zlib.MAX_WBITS)
                elif 'br' in content_encoding:
                    compressed_data = response.content
                    print('content_encoding:', content_encoding)
                    decompressed_data = brotli.decompress(compressed_data)
                else:
                    decompressed_data = response.text
            else:
                decompressed_data = response.text
        except Exception as e:  # To handle if the response is not compressed and still ended up here
            self.logger.error(f"Error decoding compressed response: {e}. Returning raw response.")
            decompressed_data = response.text
        return decompressed_data

    def get_headers(self):
        if self.simulate_browser:
            return self.new_browser_headers()
        return None

    def new_browser_headers(self) -> dict:
        """
        Generates typical Chrome browser headers
        """
        ua = UserAgent()
        return {
            'User-Agent': str(ua.chrome),
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.9,sv;q=0.8,no;q=0.7,da;q=0.6,fi;q=0.5',
            'Accept-Encoding': 'gzip, deflate, br, identity',
            'Connection': 'keep-alive',
            'Upgrade-Insecure-Requests': '1',
            'Referer': 'https://www.google.com/',
        }

    def run(self):
        """
        Starts a crawler asynchronously on the URLs loaded into the crawler (self.urls)
        """
        if self.backend == Backend.SCRAPY:
            run_globy_spider(self.urls, self.html_output_folder, self.get_headers())
        else:
            asyncio.run(self.crawl())

    def debug_latest_response(self):
        """
        Usage example:
        >>> c = WebSiteDataCrawler(debug=True)
        >>> c.load_urls_from_file("input-urls.txt")
        >>> w = c.debug_latest_response()
        >>> len(w.html_content)
        """
        logger = Logger()
        logger.set_log_level(3)
        w = None
        try:
            w = Website(self._latest_url)
            # Manually store the response in the website object
            w.html_content = self._latest_response

            w.parse_response()
            w.get_website_info()
            w.to_df()
        except Exception as e:
            self.logger.error(f"Errors: {e}")
        return w

    def load_urls_from_file(self, filename) -> None:
        with open(filename, "r") as f:
            self.urls = f.readlines()

    def load_urls_from_list(self, urls) -> None:
        self.urls = urls

    def store_dataframes_to_csv(self):
        return pd.concat(self.dataframes).sort_values(by=["url"]).to_csv(self.output_file)

    def store_html_content_to_file(self, url, html_content):
        with open(f"{self.html_output_folder}/{domain_from_url(url)}.html", "w") as f:
            f.write(html_content)

    async def get_dummy_urls(self) -> list:
        dummy_urls = [
            "https://example.com",
            "https://example.org",
            "https://example.net",
        ]
        return dummy_urls

    def print_waiting_coroutines(self, urls):
        print("Queued requests for the following URLs:")
        urls_string = ""
        for url in urls:
            urls_string = "".join([urls_string, url])
        print(urls_string)

    async def print_wait_for_coro(self, job, ts):
        url, coro = job
        while True:
            if not coro.done() and time() - ts >= 10:
                print(f"Still waiting on responses from {url} ...")
                await asyncio.sleep(5)
            elif coro.done():
                break
            await asyncio.sleep(1)

    async def fetch_url(self, url):
        """
        Fetches a URL asynchronously
        """
        if len(url) < 10:
            return url, ""

        async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=self.crawl_timeout)) as session:
            try:
                headers = self.get_headers()
                async with session.get(url, headers=headers) as response:
                    await asyncio.sleep(0.01)
                    r = await response.read()
                    if not response.ok:
                        self.logger.error(f"NOK status code on host {url}: {r.status_code}")
                        return url, ""
            except asyncio.TimeoutError as e:
                self.logger.error(f"Timeout connecting to host: {url}. {e}")
                return url, ""
            except Exception as e:
                self.logger.error(f"Error connecting to host: {url}. {e}")
                return url, ""
            else:
                if not session.auto_decompress:
                    resp = type('Request', (object,),
                                {
                                    'headers': response.headers,
                                    'content': r,
                                    'text': r.decode('utf-8'),
                                })()
                    r = self.decode_compressed_response(resp)
                else:
                    r = r.decode('utf-8')
                return url, r

    def fetch_url_sync(self, url):
        """
        Fetches a URL synchronously
        """
        if len(url) < 10:
            return url, ""
        try:
            headers = self.get_headers()
            r = requests.get(url, timeout=self.crawl_timeout, headers=headers)
        except requests.exceptions.ConnectionError as e:
            self.logger.error(f"Error connecting to host: {url}. {e}")
            return url, ""
        except requests.exceptions.Timeout as e:
            self.logger.error(f"Timeout connecting to host: {url}. {e}")
            return url, ""
        if not r.ok:
            self.logger.error(f"NOK status code on host {url}: {r.status_code}")
        text = self.decode_compressed_response(r)
        return url, text
    
    def get_website_sync(self, url) -> Website:
        w = Website(url)
        url, response = self.fetch_url_sync(url)
        w.read_response_sync(response)
        return w

    async def crawl(self):
        """
        Main function that runs the crawler
        """
        tasks, waiters = [], []
        for url in self.urls:
            task, ts = asyncio.create_task(self.fetch_url(url)), time()
            tasks.append(task)
            waiters.append(
                asyncio.create_task(self.print_wait_for_coro((url, task), ts))
            )
        self.print_waiting_coroutines(urls=self.urls)

        # Fetch and process responses as they come in
        for task in asyncio.as_completed(tasks):
            try:
                url, response = await task
            except aiohttp.client_exceptions.ClientConnectorError as e:
                self.logger.error(f"Error connecting to host: {e}. Skipping it!")
                continue
            w = Website(url)
            await w.read_response(response)
            self._latest_url, self._latest_response = url, response
            if self.debug:
                self.debug_latest_response()
            self.all_websites.append(w)
            self.dataframes.append(w.to_df())
            if self.store_html_to_file and w.ok:
                self.store_html_content_to_file(url, response)

        for waiter in asyncio.as_completed(waiters):
            await waiter
        print(f"Done crawling {len(tasks)} URLs!")
        if self.output_file is not None:
            print(f"csv output stored in {self.output_file}")
        if self.store_html_to_file:
            print(f"HTML content stored in {self.html_output_folder}/")
