from bs4 import BeautifulSoup
import pandas as pd
from time import time
import json
import re
from globy_webcrawlers.common import Logger

class Website:
    """
    Created for each website we analyze
    """

    def __init__(self, url) -> None:
        self.logger = Logger().logger
        self.url = url
        self.html_content = str()
        self.soup = None
        self.features = dict()

        """
        Will be set to True if the response is defined for the website with a non-error status code
        Note that this will always be False if Website is manually instantiated separately from the request, such as when debugging.
        """
        self.ok = False

    def add_features(self, features):
        if len(self.features) == 0:
            self.features = features
            self.features["category"] = "X"
        else:
            for key, value in features.items():
                self.features[key] = value

    def to_df(self) -> pd.DataFrame:
        try:
            df = pd.DataFrame.from_dict(self.features, orient="index").transpose()
        except Exception as e:
            self.logger.fatal(
                f"Error converting features to dataframe from url {self.url}: {e}"
            )
        else:
            return df

    def to_html(self) -> str:
        return self.soup.prettify()

    def parse_response(self) -> None:
        """
        Parses the response from a website
        """
        self.logger.info(f"Reading response from {self.url}")
        if len(self.html_content) < 100:
            return
        self.soup = BeautifulSoup(self.html_content, "html.parser")

    def is_confusing_word(self, word):
        if re.search(r'_|/|END|SLIDER|portfolio', word, re.I):
            return True
        elif len(word) < 3:
            return True
        elif word == "":
            return True
        elif word.isdigit():
            return True
        return False
    
    def new_row_object(self) -> dict:
        return {
                'text_contents': [],
                'text_lengths': [],
            }

    def new_text_tags_object(self) -> dict:
        return {
            'h1': self.new_row_object(),
            'h2': self.new_row_object(),
            'h3': self.new_row_object(),
            'h4': self.new_row_object(),
            'h5': self.new_row_object(),
            'h6': self.new_row_object(),
            'p': self.new_row_object(),
            'li': self.new_row_object()
            }

    def is_one_pager(self):

        # Check for smooth scrolling behavior
        if self.soup.find_all('a', {'data-scroll': True}):
            return True

        # Check for URL fragment navigation
        if self.soup.find_all('a', href=lambda x: x and '#' in x):
            return True

        # Check for a concise HTML structure
        body = self.soup.find('body')
        if body and len(body.find_all(recursive=False)) <= 1:
            return True

        return False

    async def read_response(self, response) -> None:
        """
        Reads the response from a website
        """
        self.html_content = response
        if len(response) < 100:
            return
        try:
            self.parse_response()
        except Exception as e:
            self.logger.error(f"Error parsing response from {self.url}: {e}")
            return
        self.ok = True

        # Extract features from the website
        self.get_website_info()
        self.get_wordpress_info()
        self.get_layout_info()
        self.add_features({"is_one_pager": self.is_one_pager()})

    def read_response_sync(self, response) -> None:
        """
        Reads the response from a website
        """
        self.html_content = response
        if len(response) < 100:
            return
        try:
            self.parse_response()
        except Exception as e:
            self.logger.error(f"Error parsing response from {self.url}: {e}")
            return
        self.ok = True

        # Extract features from the website
        self.get_website_info()
        self.get_wordpress_info()
        self.get_layout_info()
        self.add_features({"is_one_pager": self.is_one_pager()})

    def get_website_info(self) -> None:
        """
        Extracts generic properties from the website
        """
        # Number of Images
        num_images = len(self.soup.find_all("img"))
        self.logger.info(f"Number of Images: {num_images}")

        # Menu Items
        menu_items = len(self.soup.find_all(["nav", "ul", "li"]))
        self.logger.info(f"Number of Menu Items: {menu_items}")

        # Text Content
        paragraphs = self.soup.find_all("p")
        text_content = " ".join([p.text for p in paragraphs])
        self.logger.info(f"Text Content Length: {len(text_content)}")
        self.logger.debug(f"Text Content: {text_content}")

        # Metadata - Title and Description
        title = self.soup.title.text if self.soup.title else "N/A"
        description = self.soup.find("meta", {"name": "description"})
        description_content = description.get("content") if description else "N/A"
        self.logger.info(f"Title: {title}")
        self.logger.info(f"Description: {description_content}")

        # Styles and Themes
        css_links = self.soup.find_all("link", {"rel": "stylesheet"})
        stylesheets = [link.get("href") for link in css_links]
        self.logger.debug(f"Stylesheets: {stylesheets}")

        # Responsive Design
        responsive_design = (
            True if self.soup.find("meta", {"name": "viewport"}) else False
        )
        self.logger.info(f"Responsive Design: {responsive_design}")

        # Links - Internal and External
        internal_links = [
            link["href"]
            for link in self.soup.find_all("a", href=True)
            if self.url in link["href"]
        ]
        external_links = [
            link["href"]
            for link in self.soup.find_all("a", href=True)
            if self.url not in link["href"]
        ]
        self.logger.debug(
            f"Internal Links: {internal_links}, External Links: {external_links}"
        )

        # Forms
        forms = len(self.soup.find_all("form"))
        form_elements = len(self.soup.find_all(["input", "textarea", "select"]))
        self.logger.info(
            f"Number of Forms: {forms}, Number of Form Elements: {form_elements}"
        )

        # Social Media Integration
        social_media_icons = len(
            self.soup.find_all(["i", "span"], class_="social-icon")
        )
        self.logger.info(f"Number of Social Media Icons: {social_media_icons}")

        # Multimedia Embeds (YouTube Videos)
        youtube_videos = len(
            self.soup.find_all("iframe", {"src": "https://www.youtube.com/embed/"})
        )
        self.logger.info(f"Number of YouTube Videos: {youtube_videos}")

        # Store everything in a dictionary
        self.add_features(
            {
                "url": self.url,
                "num_images": num_images,
                "menu_items": menu_items,
                "text_content": len(text_content),
                "title_length": len(title),
                "description_length": len(description_content),
                "num_stylesheets": len(stylesheets),
                "responsive_design": responsive_design,
                "num_internal_links": len(internal_links),
                "num_external_links": len(external_links),
                "forms": forms,
                "form_elements": form_elements,
                "social_media_icons": social_media_icons,
                "youtube_videos": youtube_videos,
            }
        )

    # Additional features for WordPress sites
    def get_wordpress_info(self) -> None:
        """
        Extracts WordPress specific properties from the website
        """
        # WordPress Version
        wordpress_version = self.soup.find("meta", {"name": "generator"})
        wordpress_version = (
            wordpress_version.get("content") if wordpress_version else "N/A"
        )
        self.logger.info(f"WordPress Version: {wordpress_version}")

        # WordPress Theme
        wordpress_theme = self.soup.find("meta", {"name": "theme-name"})
        wordpress_theme = wordpress_theme.get("content") if wordpress_theme else "N/A"
        self.logger.debug(f"WordPress Theme: {wordpress_theme}")

        wpb_rows = self.soup.find_all('div', class_='wpb_row')
        self.logger.info(f"Number of WPB rows: {len(wpb_rows)}")

        # Store headings and paragraphs in lists
        confusing_words, tags = [], self.new_text_tags_object()
        if len(wpb_rows) > 0:
            for element in self.soup.find_all('div', class_='wpb_row'):
                for tag in tags.keys():
                    for item in element.find_all(tag):
                        text = ' '.join([string for string in item.find_all(string=True, recursive=False)]).strip()
                        if text and not self.is_confusing_word(text):
                            tags[tag]['text_contents'].append(text)
                            tags[tag]['text_lengths'].append(len(text))
                        else:
                            confusing_words.append(text)

        # WordPress Plugins (Example: extract from script tags)
        wordpress_plugins = self.soup.find_all("script", {"src": "wp-content/plugins"})
        wordpress_plugins = [plugin["src"] for plugin in wordpress_plugins]
        self.logger.debug(f"WordPress Plugins: {wordpress_plugins}")

        # Custom Post Types
        custom_post_types = self.soup.find_all(
            "script", {"type": "application/ld+json"}
        )
        try:
            custom_post_types = [json.loads(script.text) for script in custom_post_types]
        except:
            custom_post_types = []
        self.logger.debug(f"Custom Post Types: {custom_post_types}")

        # Store additional features in the dictionary
        self.add_features(
            {
                "wordpress_version": wordpress_version,
                "wordpress_theme": wordpress_theme,
                "wordpress_plugins": wordpress_plugins,
                "num_custom_post_types": len(custom_post_types),
                "num_wpb_rows": len(wpb_rows),
                "num_h1": len(tags['h1']['text_contents']),
                "num_h2": len(tags['h2']['text_contents']),
                "num_h3": len(tags['h3']['text_contents']),
                "num_h4": len(tags['h4']['text_contents']),
                "num_h5": len(tags['h5']['text_contents']),
                "num_h6": len(tags['h6']['text_contents']),
                "num_p": len(tags['p']['text_contents']),
                "num_li": len(tags['li']['text_contents']),
                "h1_lengths": tags['h1']['text_lengths'],
                "h2_lengths": tags['h2']['text_lengths'],
                "h3_lengths": tags['h3']['text_lengths'],
                "h4_lengths": tags['h4']['text_lengths'],
                "h5_lengths": tags['h5']['text_lengths'],
                "h6_lengths": tags['h6']['text_lengths'],
                "p_lengths": tags['p']['text_lengths'],
                "li_lengths": tags['li']['text_lengths'],
                "num_confusing_words": len(confusing_words),
            }
        )

    # Additional features for layout analysis
    def get_layout_info(self) -> None:
        """
        Extracts layout specific properties from the website
        """
        # Number of Sections/Divs
        num_sections = len(self.soup.find_all(["div", "section", "article"]))
        self.logger.info(f"Number of Sections/Divs: {num_sections}")

        # Page Dimensions
        page_width = self.soup.find("meta", {"name": "viewport"})
        page_width = page_width.get("content") if page_width else "N/A"
        self.logger.info(f"Page Width: {page_width}")

        # Header and Footer Detection
        num_headers = len(self.soup.find_all(["header", "header"]))
        num_footers = len(self.soup.find_all("footer"))
        self.logger.info(
            f"Number of Headers: {num_headers}, Number of Footers: {num_footers}"
        )

        # Column Count (Example: count columns in main content area)
        num_columns = len(self.soup.find_all(class_=re.compile(r"\bcol-\w+\b")))
        self.logger.info(f"Number of Columns: {num_columns}")

        # Responsive Breakpoints
        media_queries = self.soup.find_all("style")
        responsive_breakpoints = [
            query.text for query in media_queries if "media" in query.text
        ]
        self.logger.debug(f"Responsive Breakpoints: {responsive_breakpoints}")

        # Use of Flexbox or Grid
        style = self.soup.find("style")
        if style is not None:
            uses_flexbox = "display: flex" in self.soup.find("style").text
            uses_grid = "display: grid" in self.soup.find("style").text
        else:
            uses_flexbox = False
            uses_grid = False
        self.logger.info(f"Uses Flexbox: {uses_flexbox}, Uses Grid: {uses_grid}")

        # Detect Page Sections (Example: hero section, feature section)
        hero_sections = len(self.soup.find_all(class_="hero"))
        feature_sections = len(self.soup.find_all(class_="feature"))
        self.logger.info(
            f"Number of Hero Sections: {hero_sections}, Number of Feature Sections: {feature_sections}"
        )

        # Inline Styles
        inline_styles = len(self.soup.find_all(style=True))
        self.logger.info(f"Number of Elements with Inline Styles: {inline_styles}")

        # Interactive Elements (Example: sliders, carousels, accordions)
        interactive_elements = len(
            self.soup.find_all(["slider", "carousel", "accordion"])
        )
        self.logger.info(f"Number of Interactive Elements: {interactive_elements}")

        # Store additional features in the dictionary
        self.add_features(
            {
                "num_sections": num_sections,
                "page_width": page_width,
                "num_headers": num_headers,
                "num_footers": num_footers,
                "num_columns": num_columns,
                "num_responsive_breakpoints": len(responsive_breakpoints),
                "uses_flexbox": uses_flexbox,
                "uses_grid": uses_grid,
                "hero_sections": hero_sections,
                "feature_sections": feature_sections,
                "inline_styles": inline_styles,
                "interactive_elements": interactive_elements,
            }
        )
